/* --------------------------------------------------------------- */
/* mbedLED.c -- LED device I/O functions                           */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* A function to effect an action on, or query,  an LED (assumed   */
/* to be board-mounted).                                           */
/*                                                                 */
/* LED numbering is arbitrary, but if there is one or more board   */
/* LEDs then one should be accessible as 'LED 1'.                  */
/*                                                                 */
/* If there is only the one LED on the board and documentation     */
/* suggests that it be called 'LED 0' then it is recommended that  */
/* both num=0 and num=1 map to the same LED.                       */
/* --------------------------------------------------------------- */


#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"

/* --------------------------------------------------------------- */
/* ledIO --  effect an I/O action on a LED                         */
/*                                                                 */
/*   num    -- number of the LED (board-defined)                   */
/*   action -- one of LED_ON, LED_OFF, LED_INVERT, or LED_QUERY    */
/*                                                                 */
/*   returns the previous value of the LED (0 or 1), or -1 if num  */
/*     or action is invalid                                        */
/*                                                                 */
/* This maps the board LEDs as follows:                            */
/*                                                                 */
/*   num=1:  LED1                                                  */
/*   num=2:  LED2                                                  */
/*   num=3:  LED3                                                  */
/*   num=4:  LED4                                                  */
/*                                                                 */
/* --------------------------------------------------------------- */
int ledIO(int num, uint action) {       // carry out LED I/O action
  uint was;                             // previous setting
  uint bit, mask;                       // bit and corresponding mask

  // all four mbed LEDs are port 1
  #define port PORT1

  switch (num) {
    case 1: bit=18; break;
    case 2: bit=20; break;
    case 3: bit=21; break;
    case 4: bit=23; break;
    default:
      return -1;                        // not 1 through 4
    }
  mask=1<<bit;

  bitQuery(port, mask, &was);
  switch (action) {
    case LED_ON:
      bitSet(port, mask, BIT_ON);
      break;
    case LED_OFF:
      bitSet(port, mask, BIT_OFF);
      break;
    case LED_INVERT:
      bitSet(port, mask, BIT_INVERT);
      break;
    case LED_QUERY:
      // no change needed
      break;
    default:
      return -1;
    }

  return was;
  } // ledIO


