/* --------------------------------------------------------------- */
/* stmVCom.c -- Virtual COM device I/O functions                   */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* The vcom functions communicate with the host computer using a   */
/* USB link.                                                       */
/*                                                                 */
/* Note that on Windows a driver may need to be installed for this */
/* to work.  This should be found automatically; if not, try the   */
/* stmvirtcom.inf driver.                                          */
/*                                                                 */
/* The primary functions here provide initialization, the usual    */
/* getc and putc, and stop.  Additional exposed functions are the  */
/* callback function for interfacing with the USB library and      */
/* 'USART' functions which are called when the data terminal       */
/* is reset or is started.                                         */
/*                                                                 */
/* For reading, up to 35 characters will be buffered here; if      */
/* the buffer fills (the host sends them faster than the           */
/* application can read them) then characters will be lost.        */
/*                                                                 */
/* Similarly, up to 320 characters will be buffered for output;    */
/* again, characters will be lost if too many are sent.  Neither   */
/* vcomPutc nor vcomGetc will block.  The host application may     */
/* also discard characters (for example, if the terminal           */
/* emulator is stopped then restarted, the USB device is not       */
/* informed of this and characters sent while the terminal is      */
/* not running are accepted but then lost).                        */
/* --------------------------------------------------------------- */
// The first version of this code took the approach of sending a
// packet to the host for every character put; this worked, but
// as a large timeout is required (in case of slow host response)
// this could slow the caller too much.
//
// The current version uses a ring of packet-sized buffers which can
// fill while data are being sent to the host.  In theory about 2000
// bytes would be needed to accept data at 10kB/s and to survive a
// 200ms glitch, but rather fewer are needed for most vcom
// applications.  The buffer size is BPUTSIZE*PACKETSIZE; BPUTSIZE
// can be increased below if required.


#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"

#include "stm32f10x.h"
#include "stm32f10x_rcc.h"
#include "usb_lib.h"          // USB library functions & equates
#include "usb_desc.h"
#include "usb_istr.h"
#include "usb_pwr.h"

// callbacks used by usb_prop [using same name to avoid code changes]
void USART_Config_Default(void);
bool USART_Config(void);

// state
static flag started=0;        // 1 if active
static flag transmitting=0;   // 1 if wait for byte to be sent
static flag bputbusy=0;       // 1 while vcomPutc fiddles
static flag terminal=0;       // 1 if host terminal is active

// vcomPutC buffer ring
#define  BPUTSIZE      5      // number of buffers in ring
#define  PACKETSIZE   64      // packet size to buffer (max. 64)
// a transmit packet and its accounting
typedef struct {
  uint count;                 // bytes pending (already filled)
  byte data[PACKETSIZE];      // holds bytes
  } packet;
static packet bput[BPUTSIZE]; // ring of buffers
static packet *bputread;      // next packet to send
static packet *bputwrite;     // packet being filled

// vcomGetC circular buffer
#define BGETSIZE 36           // (one remains unused)
static byte bget[BGETSIZE];   // input buffer
static byte *bgetread;        // next-to-read pointer
static byte *bgetwrite;       // next-to-write pointer

/* --------------------------------------------------------------- */
/* EP1_IN_Callback -- called by USB library for endpoint 1 (IN)    */
/*                                                                 */
/* Called when inbound transmission is complete.                   */
/* --------------------------------------------------------------- */
void EP1_IN_Callback(void) {
  transmitting=0;             // allow next send
  } // EP1_IN_Callback

/* --------------------------------------------------------------- */
/* EP3_OUT_Callback -- called by USB library for endpoint 3 (OUT)  */
/*                                                                 */
/* Called when a packet of bytes has been received from the host.  */
/* --------------------------------------------------------------- */
void EP3_OUT_Callback(void) {
  uint16 count;                              // number of bytes received
  byte buffer[VIRTUAL_COM_PORT_DATA_SIZE];   // work
  byte *bp=buffer;                           // buffer pointer

  count=GetEPRxCount(ENDP3);                           // bytes pending
  PMAToUserBufferCopy(buffer, ENDP3_RXADDR, count);    // make safe copy

  // add new bytes to bget (if overrun, discard)
  for (; count>0; count--, bp++) {
    byte *next=bgetwrite+1;                  // -> where *next* byte will go
    if (next==bget+BGETSIZE) next=bget;      // wrap
    if (next==bgetread) break;               // would overrun
    *bgetwrite=*bp;
    bgetwrite=next;
    }

  SetEPRxValid(ENDP3);                       // ready for next
  } // EP3_OUT_Callback

/* --------------------------------------------------------------- */
/* SOF_Callback -- called by USB library on frame start            */
/*                                                                 */
/* The device is polled by the host every 1ms; this callback is    */
/* then invoked.                                                   */
/* --------------------------------------------------------------- */
#define INIT_TIMEOUT 250
#define WAIT_TIMEOUT 250
void SOF_Callback(void) {
  static uint init=INIT_TIMEOUT;   // initial delay
  static uint waited=0;            // time waiting since last transmit
  packet *next;                    // work

  // check if there are data to send; if there are then if the
  // terminal is there and if vcomPutc is not updating pointers then
  // looks promising to start a transmit
  if (bputread->count==0           // nothing to send
   || !terminal                    // no receiver
   || bputbusy) return;            // locked out

  // also quit if transmission in progress; transmitting might never
  // get cleared if the host connection is lost; so check timeout if
  // waiting; if the timeout has expired then force through another
  // transmit -- however in practice a full reset is probably needed
  if (transmitting) {                   // busy
    waited++;
    if (waited<=WAIT_TIMEOUT) return;   // carry on waiting
    // here when timeout occurs -- for sure no terminal application
    terminal=0;
    }

  // some terminals connect before finishing screen build, so when
  // first called delay before sending first transmission
  if (init>0) {
    init--;
    return;
    }

  // (the following code could be made into a local function and then
  // also called from EP1_IN_Callback if there are bytes to send;
  // this would avoid waiting up to 1ms for the next host poll)

  // ready to transmit
  transmitting=1;                  // state
  waited=0;                        // reset timeout

  // copy byte(s) to usb packet-memory-area for end point 1
  UserToPMABufferCopy(bputread->data, ENDP1_TXADDR, bputread->count);
  SetEPTxCount(ENDP1, bputread->count);  // set count of bytes to send
  SetEPTxValid(ENDP1);             // enable endpoint for transmission

  // show the packet as empty (free for reuse)
  bputread->count=0;               // indicate data sent

  // if the next packet has data pending, the write pointer, having
  // writ the packet just sent, has moved on; step to the next packet
  // so it will be sent in due course

  next=bputread+1;
  if (next==&bput[BPUTSIZE]) next=bput; // wrap
  if (next->count!=0) bputread=next;    // step to it
  } // SOF_Callback

/* --------------------------------------------------------------- */
/* USART_Config -- called by library when terminal exists          */
/*                                                                 */
/* This is called when a 'real' host terminal device is attached.  */
/* --------------------------------------------------------------- */
bool USART_Config(void) {
  terminal=1;
  return 1;
  } // USART_Config

/* --------------------------------------------------------------- */
/* USART_Config_Default -- called by library when device started   */
/*                                                                 */
/* This is called when explicit initialization of the Virtual COM  */
/* port is requested.  This implies no terminal yet.               */
/* --------------------------------------------------------------- */
void USART_Config_Default(void) {
  terminal=0;
  } // USART_Config_Default

/* --------------------------------------------------------------- */
/* vcomGetcIO -- get character                                     */
/*                                                                 */
/*   returns character read from host, or -1 if none there         */
/*                                                                 */
/* This function does not block.                                   */
/* --------------------------------------------------------------- */
int vcomGetcIO(void) {
  byte b;                               // work

  if (bgetread==bgetwrite || !started) return -1; // nothing to send

  b=*bgetread;                          // save usable byte

  // bump read point to show byte used
  bgetread++;
  if (bgetread==bget+BGETSIZE) bgetread=bget; // wrap

  return b;
  } // vcomGetcIO

/* --------------------------------------------------------------- */
/* vcomInitIO -- initialize USB port for I/O                       */
/*                                                                 */
/*   returns 0 if OK, negative if interrupt already in use         */
/*                                                                 */
/* This does not switch a software connection as this is           */
/* device-dependent (port, pin, and direction), so the caller      */
/* should use bitSet to make the connection before calling this.   */
/* --------------------------------------------------------------- */
int vcomInitIO(void) {
  int rc, i;                  // work

  if (started) return 0;      // already active

  // set up bget and bput, and zero latter's counts
  bgetread=bget;
  bgetwrite=bget;
  bputread=bput;
  bputwrite=bput;
  for (i=0; i<BPUTSIZE; i++) bput[i].count=0;
  started=1;

  // STM setup will have set USBCLK source appropriate for 48MHz */
  // (e.g. PLLCLK_1Div5 for 72MHz system clock)
  RCC_APB1PeriphClockCmd(RCC_APB1Periph_USB, ENABLE); // enable USB clock

  // set up interrupt handler using Tollos USB priority (quite high)
  rc=sysVector(USB_LP_CAN1_RX0_IRQn, USB_Istr, SYSTEM_USBPRIO);

  USB_Init();

  return rc;
  } // vcomInitIO

/* --------------------------------------------------------------- */
/* vcomPutcIO -- send character to virtal com port                 */
/*                                                                 */
/*   ch -- character to send                                       */
/*                                                                 */
/*   returns 0 if OK, or -1 if vcom has not been initialized, or   */
/*   -2 if there is a data overrun (bytes have been queued faster  */
/*   than they could be sent to the host)                          */
/*                                                                 */
/* This function does not block.                                   */
/*                                                                 */
/* Note that this must not be called from an interrupt handler     */
/* that has a higher priority than USB.                            */
/* --------------------------------------------------------------- */
int vcomPutcIO(int ch) {
  if (!started) return -1;

  if (bputwrite->count==PACKETSIZE) {        // packet is full
    packet *next=bputwrite+1;
    if (next==&bput[BPUTSIZE]) next=bput;    // wrap
    if (next->count!=0) return -2;           // overrun
    bputwrite=next;                          // clear to use next
    }

  bputbusy=1;                 // prevent handler from meddling
  bputwrite->data[bputwrite->count]=(byte)ch;     // place data
  bputwrite->count++;                             // and account for it
  bputbusy=0;                 // release lock

  return 0;
  } // vcomPutcIO

/* --------------------------------------------------------------- */
/* vcomStopIO -- stop the device cleanly                           */
/*                                                                 */
/* This releases the interrupt handler after stopping the          */
/* device.                                                         */
/*                                                                 */
/* This does not switch a software connection as this is           */
/* device-dependent (port, pin, and direction), so the caller      */
/* should use bitSet to make the connection an input after         */
/* calling this.                                                   */
/* --------------------------------------------------------------- */
void vcomStopIO(void) {
  if (!started) return;
  PowerOff();
  sysVector(USB_LP_CAN1_RX0_IRQn, NULL, SYSTEM_USBPRIO);
  started=0;                  // indicate
  } // vcomStopIO

