/* --------------------------------------------------------------- */
/* tollosUtil.c -- useful macros and functions for Tollos          */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* The macros are defined in ""tollosUtil.h"", and are:            */
/*                                                                 */
/*   abs(x)     -- return the absolute value of x                  */
/*   iswap(x,y) -- swap the values on the ints x and y             */
/*   max(a,b)   -- return the larger of x and y                    */
/*   min(a,b)   -- return the smaller of x and y                   */
/*   signum(x)  -- return the sign of x (-1, 0, or 1)              */
/*                                                                 */
/* The functions are miscellaneous.                                */
/* --------------------------------------------------------------- */

#include "tollos.h"

/* --------------------------------------------------------------- */
/* byte2hex -- convert a byte to two hex chars                     */
/*                                                                 */
/*   b       -- byte to convert                                    */
/*   hex     -- points to where ASCII bytes should go              */
/*   upper   -- 1 to use ABCDEF, 0 for abcdef                      */
/*                                                                 */
/* This writes only the two characters; no terminating zero byte   */
/* is added.  No error is possible.                                */
/* --------------------------------------------------------------- */
void byte2hex(byte b, char *hex, flag upper) {
  static const char *uppers="0123456789ABCDEF";
  static const char *lowers="0123456789abcdef";
  const char *letters=(upper ? uppers : lowers);  // select table

  hex[0]=letters[(unsigned)b>>4];
  hex[1]=letters[b & 0x0f];
  } // byte2hex

/* --------------------------------------------------------------- */
/* fastrand -- return a random number in the range 0 to n-1        */
/*                                                                 */
/*   n    -- defines the random number range                       */
/*   seed -- static or persistent seed                             */
/*                                                                 */
/* n should be <30,000,000; the seed is provided by the caller     */
/* initially but should not then be altered by the caller.         */
/*                                                                 */
/* This uses the conventional  X(n+1)=mod((A*X(n) + C),M) as used  */
/* in the Rexx function package, but without some refinements.     */
/*                                                                 */
/* The values for A and M are from the table in D.E.Knuth's        */
/* book, "Seminumerical Algorithms", Addison-Wesley, 1981,         */
/* page 102 (section 3.3.2).                                       */
/* These particular values are attributed to Lavaux and Janssens.  */
/* * C is chosen as 1                                              */
/* * M is 2**32                                                    */
/* * A is 1664525      (RNDFACT)                                   */
/*                                                                 */
/* --------------------------------------------------------------- */
int fastrand(uint n, uint *seed) {
  *seed=(*seed)*1664525L+1;        // don't care about overflow
  return (int)((*seed>>3)%n);      // return 0->N-1
  } // fastrand


/* --------------------------------------------------------------- */
/* hex2byte -- convert two hexadecimal characters to a byte        */
/*                                                                 */
/*   hex     -- points to two ASCII bytes                          */
/*   b       -- points to byte which will receive the converted    */
/*              hexadecimal; set to 0x00 if error                  */
/*   returns 0 if OK, -1 if error                                  */
/*                                                                 */
/* This accepts both uppercase and lowercase A->F and a->f.        */
/* --------------------------------------------------------------- */
int hex2byte(char *hex, byte *b) {
  char *end;        // end char +1
  *b=0;
  for (end=hex+2; hex<end; hex++) {
    *b<<=4;
    if (*hex>='0' && *hex<='9') *b|=(*hex-'0');
    else if (*hex>='A' && *hex<='F') *b|=(*hex-'A'+10);
    else if (*hex>='a' && *hex<='f') *b|=(*hex-'a'+10);
    else {*b=0x00; return -1;}
    }
  return 0;
  } // hex2byte

/* --------------------------------------------------------------- */
/* isqrtg -- return the square root of an int, with guess          */
/*                                                                 */
/*   num   -- the number to square root                            */
/*   guess -- is the number to converge from (e.g., last similar   */
/*            result); if < 1, isqrtg will make the first guess    */
/*                                                                 */
/*   returns the integer square root of num, such that             */
/*     result*result < num; if num is less than zero then 0 is     */
/*     returned                                                    */
/*                                                                 */
/* Note: the maximum input value is 2**31-1 (2147483647)           */
/* hence the maximum possible returned value is 46340.             */
/* --------------------------------------------------------------- */
// mfc, derived from PalmGlobe lsqrtg
int isqrtg(int num, int guess) {
  int new, old;

  if (num<9) {                /* special case low values for safety */
    if (num<=0) return 0;     /* special case + safety */
    if (num< 4) return 1;     /* 1, 2, 3 */
    return 2;}                /* 4-8 */

  /* Note: first guess has dramatic performance impact -- a value   */
  /* 1, or N/16, both doubled total GLOBE draw time.                */
  /* N/1024+1, and 32,000 both worked well.  Dynamic better.        */
  /* dynamically calculated first guess, which is (just) better.    */
  old=guess;                  /* The first guess is a given */
  if (old<=0) old=32001;
  for (;;) {                  /* forever */
    new=(signed int)((unsigned int)(num/old+old)>>1); /* new guess */
    if ((new-old)<2
     && (new-old)>-2) break;  /* Done! */
    old=new;
    }
  return new;
  } // isqrtg

/* --------------------------------------------------------------- */
/* i2str -- layout integer                                         */
/*                                                                 */
/*   string -- where to layout the integer                         */
/*   num    -- the integer to lay out (not 0x80000000)             */
/*   len    -- width of the field (string) to lay out into         */
/*                                                                 */
/*   returns 0 if OK, 1 if overflow (does not fit)                 */
/*                                                                 */
/* This writes num right-aligned and signed in the string field.   */
/* No terminating 0 is written.                                    */
/* --------------------------------------------------------------- */
int i2str(char *string, int num, int len) {
  char *c=string+len-1;                 // -> rightmost char
  char sign='+';
  int i;
  if (num<0) {num=-num; sign='-';}      // broken if 0x80000000
  for (i=0; i<len; i++) string[i]=' ';  // clear
  for (;;) {
    if (c==string) {                    // no room for sign
      for (i=0; i<len; i++) string[i]='*'; // show
      return 1;
      }
    *c=(num%10)+'0';
    num=num/10;
    c--;
    if (num==0) break;
    }
  *c=sign;
  return 0;
  } // i2str

/* --------------------------------------------------------------- */
/* memcmp -- compare bytes as per C library                        */
/* --------------------------------------------------------------- */
int memcmp(const void *s1, const void *s2, const size_t n) {
  uint u;
  const char *p1=s1;
  const char *p2=s2;
  for (u=0; u<n; u++, p1++, p2++) {
    if (*p1 < *p2) return -1;
    if (*p1 > *p2) return +1;
    }
  return 0;    // all bytes equal
  } // memcmp

/* --------------------------------------------------------------- */
/* memcpy -- copy bytes as per C library                           */
/* --------------------------------------------------------------- */
void *memcpy(void *s1, const void *s2, size_t n) {
  uint u;
  char *p1=s1;
  const char *p2=s2;
  for (u=0; u<n; u++, p1++, p2++) *p1=*p2;
  return s1;
  } // memcpy

/* --------------------------------------------------------------- */
/* memset -- set bytes as per C library                            */
/* --------------------------------------------------------------- */
void *memset(void *s, const int c, size_t n) {
  uint u;
  char *p=s;
  for (u=0; u<n; u++, p++) *p=c;
  return s;
  } // memset

/* --------------------------------------------------------------- */
/* nop -- do nothing with an int                                   */
/*                                                                 */
/* This is used for placating compilers when parameters are        */
/* intentionally or temporarily unused.                            */
/* --------------------------------------------------------------- */
int nop(int i) {
  return i;
  } // nop

/* --------------------------------------------------------------- */
/* streq -- caseless string compare                                */
/*                                                                 */
/*   char1 -- first string to compare                              */
/*   char2 -- second string                                        */
/*                                                                 */
/*   returns 1 if the strings caseless-compare equal, 0 otherwise  */
/*                                                                 */
/* --------------------------------------------------------------- */
int streq(const char *char1, const char *char2) {
  for (;;char1++, char2++) {
    if (*char1==*char2) {if (*char1=='\0') return 1;}
     else if (toupper(*char1)!=toupper(*char2)) return 0;
    } // forever
  } // streq

/* --------------------------------------------------------------- */
/* strf -- a mini printf to send a string                          */
/*                                                                 */
/*   putC   -- the function to which to send formatted characters  */
/*   format -- format and data string                              */
/*   args   -- list of other arguments as required                 */
/*                                                                 */
/*   returns the number (count) of characters sent                 */
/*                                                                 */
/* The following format indicators are supported:                  */
/*                                                                 */
/*   %%   -- '%'                                                   */
/*   %c   -- single char                                           */
/*   %d,i -- int                                                   */
/*   %m   -- minidec decimal number  [iff MINIDEC is defined]      */
/*   %s   -- string                                                */
/*   %u   -- unsigned int                                          */
/*   %x   -- hexadecimal                                           */
/*   %\n  -- LF only (otherwise ""\n"" becomes CRLF)               */
/*                                                                 */
/* The four ""printf"" format flags (""+ - 0 #"") are allowed but  */
/* are ignored, except:                                            */
/*   '+' on integers and decimals (requests + signs)               */
/*   '0' on 'x' (e.g., ""%08x"" for an 8-hexadecimal-digit result) */
/*   '#' on 'x' (prefixes '0x')                                    */
/*                                                                 */
/* The 'h', 'l', and 'L' modifiers are not supported.              */
/*                                                                 */
/* Field width is ignored except for 'x' (where it is capped to 8) */
/* and decimals ('m'); the latter will fill with "****" if the     */
/* number does not fit in the specified field width.               */
/*                                                                 */
/* Field precision (e.g., %5.7) is not supported; for decimals,    */
/* use Quantize on a number to set its 'precision'.                */
/*                                                                 */
/* A ""\n"" in the string is expanded to ""\r\n"" unless preceded  */
/* by ""%"".                                                       */
/* Almost no error checking is carried out.                        */
/* --------------------------------------------------------------- */
int strf(putter putC, const char *format, va_list args) {
  const char *f;         // pointer in format string
  int count=0;           // chars put counter
  uint width=0;          // field width [0 means none]
  flag sign;             // 1 if '+' flag
  flag neg;              // 1 if '-' flag
  flag zero;             // 1 if '0' flag
  flag alt;              // 1 if '#' flag

  for (f=format; *f!=0; f++) {
    if (*f!='%') {
      if (*f=='\n') {putC('\r'); count++;}
      putC(*f); count++;
      continue;
      }
    // have %
    f++;
    if (*f=='%') {putC(*f); count++; continue;}

    neg=0; sign=0; alt=0; zero=0;
    for (;; f++) {           // handle leading flags in any order
      if (*f=='-') neg=1;
      else if (*f=='+') sign=1;
      else if (*f=='0') zero=1;    // (strips leading zeros)
      else if (*f=='#') alt=1;
      else break;
      } // flags loop
    width=0;                       // gather width, if any
    for (; *f>='0' && *f<='9'; f++) width=width*10+(*f-'0');
    // [dot and precision would be gathered here]

    switch(*f) {                   // format indicator
      case 'c': {
        char c=(char)va_arg(args, int);
        putC(c); count++;
        break;}

      case 'u':
      case 'd':
      case 'i': {
        uint u=(uint)va_arg(args, uint);
        uint p;          // powers of 10
        flag had;        // had leading digit
        if (*f!='u') {   // signed
          int i=(int)u;
          if (i<0) {
            putC('-'); count++;
            if (u==0x80000000) u=2147483648U;
             else u=-i;
            }
           else if (sign) {putC('+'); count++;}
          }
         else if (sign) {putC('+'); count++;}
        // u is now abs(arg), such that 0 <= u <= 4294967295
        had=0;
        if (u==0)  {putC('0'); count++;}
         else for (p=1000000000; p; p/=10) {
          uint d=u/p;
          if (d) had=1;  // had leading digit
          u%=p;
          if (had) {putC('0'+d); count++;}
          }
        break;}

      #ifdef MINIDEC
      case 'm': {
        dec num=(dec)va_arg(args, dec);

        char buff[20];   // work, >17
        char *cp=buff;   // work

        if (sign && !decIsNegative(num)) {buff[0]='+'; cp++;}
        decToString(num, cp);

        if (width!=0) {
          int len=strlen(buff);
          int pad=width-len;
          if (pad<0) {
            for (; width>0; width--) {putC('*'); count++;} // won't fit
            break;}
          for (; pad>0; pad--) {putC(' '); count++;}
          }
        for (cp=buff; *cp; cp++) {putC(*cp); count++;}
        break;}
      #endif

      case 's': {
        char *cp=(char *)va_arg(args, int);
        if (cp!=NULL) for (; *cp; cp++) {putC(*cp); count++;}
        break;}

      case 'x':
      case 'X': {
        uint u=(uint)va_arg(args, uint);
        byte upper=(*f=='X');
        flag had;
        char x[8];
        int i;
        // convert to hex
        for (i=6; i>=0; i-=2) {
          byte2hex(u & 0xff, &x[i], upper);
          u>>=8;
          }
        if (width==0 || width>8) width=8;
        if (alt) {putC('0'); count++; putC(*f); count++;}
        had=0;
        for (i=8-width; i<8; i++) {
          if (!zero && !had && x[i]=='0') continue;    // skip leading 0s
          had=1;
          putC(x[i]); count++;
          }
        break;}


      default: {
        putC(*f); count++;
        break;}
      } // format char

    } // format scan

  return count;
  } // strf

